// Class QWindow 
// Widget which is a GUI-Window.
// Filename QWindow.cpp
// FaPra 2003-2004, A5,  Alexander Kramer


#include "QWindow.h"
#include <iostream>
using namespace std;

// ----------------------------------------------------------

#include "QWindow_Handlers.h"
//! draws frame
void QWindow::drawFrame()
{

  const int border = TitleBarBorder;

  int fleft = 0;
  int ftop  = 0;

  int fwidth = Width;
  int fheight = Height;

  // draw client region - fill with the background color
  LookAndFeel->getBackGroundColor()->setAsCurrent();
  glBegin(GL_POLYGON);
  LookAndFeel->drawRect(fleft,ftop,Width,Height);
  glEnd();

  // draw up bevel and down bevel to simulate round frame
  LookAndFeel->drawBevel(fleft,ftop,fwidth,fheight,bsUp);
  LookAndFeel->drawBevel(fleft+border,ftop+border,fwidth-2*border,fheight-2*border,bsDown);

  // check if window is in focus path. If so sets color of the title bar and font to bold
  // (Highlighting of the active window)
  if (Owner->getFocusPath() == this)
   {
     TitleBarColor->setAsCurrent();
     Font->setBold(true);
   }
  else
   {
     TitleBarColorInactive->setAsCurrent();
     Font->setBold(false);
   }

  // draw window title bar
  glBegin(GL_POLYGON);
  LookAndFeel->drawRect(fleft+border,ftop+border,fwidth-2*border,TitleBarHeight);
  glEnd();

  // draw window title bar caption
  int maxLen = Font->getLengthForWidth(Caption.c_str(),MaximizeButton->getLeft()-(fleft+border+2));
  Font->drawAtPos(Caption.c_str(),fleft+border+2,ftop+border+2,maxLen);

  // sets font to normal
  Font->setBold(false);

  // If window is resizable, draw a kind of triangle in the right bottom corner
  if (Resizable)
  {
     glColor3f(1,1,1);
     glBegin(GL_LINES);
     glVertex2i(fwidth-border,fheight-12);
     glVertex2i(fwidth-12,fheight-border);

     glVertex2i(fwidth-border,fheight-8);
     glVertex2i(fwidth-8,fheight-border);

     glEnd();
  }


  // check if window is currently resizing or dragging and draw nice looking info about position/size
  if (Resizing)
  {
     char buf[101];
     snprintf(buf,100,"resize to %dx%d",Width,Height);
     drawCoordInfos(fwidth+10,fheight-10,buf);
  }
  else
  if (Dragging)
  {
     char buf[101];
     snprintf(buf,100,"move to (%d,%d)",Left,Top);
     drawCoordInfos(fleft+5,ftop-25,buf);
  }

} // end of drawFrame()

//! draws info about size/position. Semi transparent rectangle with window font 
void QWindow::drawCoordInfos( int  x , int  y , const  char*  Text )
{
  float textwidth = Font->getTextWidth(Text);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(0.2,0.2,0.2,0.5);
  glBegin(GL_POLYGON);
  LookAndFeel->drawRect(x-5,y-2,textwidth+10,Font->getSize()+5);
  glEnd();
  glDisable(GL_BLEND);
  Font->drawAtPos(Text,x,y,-1);
} // end of drawCoordInfos()

//! draws close button
void QWindow::drawCloseButton()
{

  // need this to modulate the texture
  glColor4f(1,1,1,1);

  LookAndFeel->drawCloseButton(CloseButton->getState() == bsDown,
     CloseButton->getWidth(), CloseButton->getHeight());
} // end of drawCloseButton()

//! draws maximize button
void QWindow::drawMaximizeButton()
{

  // need this to modulate the texture
  glColor4f(1,1,1,1);

  LookAndFeel->drawMaximizeButton(MaximizeButton->getState() == bsDown,
    MaximizeButton->getWidth(),  MaximizeButton->getHeight());
} // end of drawMaximizeButton()

//! maximize / restore window. If window is maximized, restore its size/position, else maximize its size 
void QWindow::maximize()
{
  if (Maximized)
  {
     setLeft(OldLeft);
     setTop(OldTop);
     setWidth(OldWidth);
     setHeight(OldHeight);
     Maximized = false;
  }
  else
  {
    OldLeft = Left;
    OldTop = Top;
    OldWidth = Width;
    OldHeight = Height;
    setLeft(0);
    setTop(0);
    setWidth(Owner->getWidth() - Owner->getClientBorderLeft() - Owner->getClientBorderRight());
    setHeight(Owner->getHeight() - Owner->getClientBorderTop() - Owner->getClientBorderBottom());
    Maximized = true;
  }
} // end of maximize()

//! draws itself with OpenGL. Here is the OpenGL drawing code
void QWindow::draw( QEvent*  e )
{
   drawFrame();

} // end of draw()

//! layout the children widgets. This method is called every time the size is changed. Have to be overwritten for respond on size changes
void QWindow::layout()
{
  CloseButton->setLeft(Width - CloseButton->getWidth()-TitleBarBorder-3);
  MaximizeButton->setLeft(CloseButton->getLeft() - MaximizeButton->getWidth()-3);
} // end of layout()

//! have to return true if widget should be dragged with mouse. This method is overwritten in sub classes. If returns true, widget will be dragged
bool QWindow::canDrag( QEvent*  e )
{

 return   (e->getY() - Top) < TitleBarHeight; 
} // end of canDrag()

//! processing mouse/keyboard events which are in widgets region. Mouse coordinates are relative to Owner
bool QWindow::processEvent( QEvent*  e )
{
      
 if ( Resizable &&
     (e->getButton() == mbLeft) &&
     (e->getType() == etMouseDown) &&
     ((Width + Left - e->getX()) < 10) &&
     ((Height + Top - e->getY()) < 10)
   )
 {
   ResizeStartX = (Width - e->getX());
   ResizeStartY = (Height - e->getY());
   Resizing = true;
   return true;
 }
 return false;
} // end of processEvent()

//! processing mouse events which are in Owner widget region. Mouse coordinates are relative to Owner
bool QWindow::processMouseOwner( QEvent*  e )
{

 if (
      (e->getButton() == mbLeft) &&
      (e->getType() == etMouseMotion)
     )
 {
   if (Resizing)
   {

      int newWidth = ResizeStartX + e->getX();
      int newHeight = ResizeStartY + e->getY();

      if (newWidth < MinResizeWidth) newWidth = MinResizeWidth;
      if (newHeight < MinResizeHeight) newHeight = MinResizeHeight;

      if (Owner)
      {
         if (newWidth < Owner->getWidth() - Left)
           setWidth(newWidth);
         else
           setWidth( Owner->getWidth() - Left);

         if (newHeight < Owner->getHeight() - Top)
           setHeight(newHeight);
         else
           setHeight( Owner->getHeight() - Top);
      }
      else
      {
         setWidth(newWidth);
         setHeight(newHeight);
      }

   }
 }
 else
  Resizing = false;

 return false;
} // end of processMouseOwner()

//! called if a child is added/removed
void QWindow::childrenChanged()
{
} // end of childrenChanged()

//! defines whether control can have keyboard focus. Should return true if Widget wants to get keyboard focus
bool QWindow::canFocus( QEvent*  e )
{
 return true;
} // end of canFocus()

//! initialization code
void QWindow::Init()
{
  Caption = "Window";

  TitleBarColor->set(0.0, 0.0, 0.9, 1.0);
  TitleBarColorInactive->set(0.2,0.2,0.25,1.0);

  Font->getColor()->set(1,1,1,1);
  Font->setSize(16);


  CloseButton->onClick = CloseButtonHandler;
  CloseButton->onCustomDraw = drawCloseButtonHandler;
  CloseButton->setHeight(TitleBarHeight-4);
  CloseButton->setWidth (CloseButton->getHeight());
  CloseButton->setTop(TitleBarBorder+2);

  MaximizeButton->onClick = MaximizeButtonHandler;
  MaximizeButton->onCustomDraw = drawMaximizeButtonHandler;
  MaximizeButton->setHeight(CloseButton->getHeight());
  MaximizeButton->setWidth(CloseButton->getWidth());
  MaximizeButton->setTop(CloseButton->getTop());

  ClientBorderTop = TitleBarBorder+TitleBarHeight+2;
  ClientBorderLeft = TitleBarBorder+1;
  ClientBorderRight = TitleBarBorder+1;
  ClientBorderBottom = TitleBarBorder+1;

  MinWidth = MinResizeWidth;
  MinHeight = MinResizeHeight;

} // end of Init()

//! clean up code
void QWindow::CleanUp()
{

} // end of CleanUp()

//! getter method for TitleBarHeight
int QWindow::getTitleBarHeight()
{
return TitleBarHeight;
} // end of getTitleBarHeight()

//! setter method for TitleBarHeight
void QWindow::setTitleBarHeight( int  newValue )
{
TitleBarHeight = newValue;
} // end of setTitleBarHeight()

//! getter method for TitleBarBorder
int QWindow::getTitleBarBorder()
{
return TitleBarBorder;
} // end of getTitleBarBorder()

//! setter method for TitleBarBorder
void QWindow::setTitleBarBorder( int  newValue )
{
TitleBarBorder = newValue;
} // end of setTitleBarBorder()

//! getter method for Caption
string QWindow::getCaption()
{
return Caption;
} // end of getCaption()

//! setter method for Caption
void QWindow::setCaption( string  newValue )
{
Caption = newValue;
} // end of setCaption()

//! getter method for MinResizeWidth
int QWindow::getMinResizeWidth()
{
return MinResizeWidth;
} // end of getMinResizeWidth()

//! setter method for MinResizeWidth
void QWindow::setMinResizeWidth( int  newValue )
{
MinResizeWidth = newValue;
} // end of setMinResizeWidth()

//! getter method for MinResizeHeight
int QWindow::getMinResizeHeight()
{
return MinResizeHeight;
} // end of getMinResizeHeight()

//! setter method for MinResizeHeight
void QWindow::setMinResizeHeight( int  newValue )
{
MinResizeHeight = newValue;
} // end of setMinResizeHeight()

//! getter method for Maximized
bool QWindow::getMaximized()
{
return Maximized;
} // end of getMaximized()

//! setter method for Maximized
void QWindow::setMaximized( bool  newValue )
{
Maximized = newValue;
} // end of setMaximized()

//! getter method for Resizable
bool QWindow::getResizable()
{
return Resizable;
} // end of getResizable()

//! setter method for Resizable
void QWindow::setResizable( bool  newValue )
{
Resizable = newValue;
} // end of setResizable()

//! getter method for CloseButton
QButton* QWindow::getCloseButton()
{
return CloseButton;
} // end of getCloseButton()

//! getter method for MaximizeButton
QButton* QWindow::getMaximizeButton()
{
return MaximizeButton;
} // end of getMaximizeButton()

//! getter method for Font
QFont* QWindow::getFont()
{
return Font;
} // end of getFont()

//! getter method for TitleBarColor
QColor* QWindow::getTitleBarColor()
{
return TitleBarColor;
} // end of getTitleBarColor()

//! getter method for TitleBarColorInactive
QColor* QWindow::getTitleBarColorInactive()
{
return TitleBarColorInactive;
} // end of getTitleBarColorInactive()

// WARNING : The implementation of this method will be automatically generated with code generator.
// To prevent this add '//KEEP' at the first line of the implementation.(first line after '{')
//! prints itself to the cout
void QWindow::print()
{
cout << "Class : QWindow" << endl;
cout << "TitleBarHeight = " << TitleBarHeight << endl;
cout << "TitleBarBorder = " << TitleBarBorder << endl;
cout << "Caption = " << Caption << endl;
cout << "MinResizeWidth = " << MinResizeWidth << endl;
cout << "MinResizeHeight = " << MinResizeHeight << endl;
cout << "Maximized = " << Maximized << endl;
cout << "Resizable = " << Resizable << endl;
cout << "CloseButton = " << CloseButton << endl;
cout << "MaximizeButton = " << MaximizeButton << endl;
} // end of print()

// WARNING : The implementation of this method will be automatically generated with code generator.
// To prevent this add '//KEEP' at the first line of the implementation.(first line after '{')
//! Constructor
 QWindow::QWindow( QWidget*  aOwner ):QWidget(aOwner)
{
Resizing = false;
OldLeft = 0;
OldTop = 0;
OldWidth = 0;
OldHeight = 0;
ResizeStartX = 0;
ResizeStartY = 0;
TitleBarHeight = 23;
TitleBarBorder = 2;
MinResizeWidth = 60;
MinResizeHeight = 40;
Maximized = false;
Resizable = true;
CloseButton = new QButton(this);
MaximizeButton = new QButton(this);
Font = new QFont;
TitleBarColor = new QColor;
TitleBarColorInactive = new QColor;
Init();
layout();
} // end of QWindow()

// WARNING : The implementation of this method will be automatically generated with code generator.
// To prevent this add '//KEEP' at the first line of the implementation.(first line after '{')
//! Destructor
 QWindow::~QWindow()
{
delete TitleBarColorInactive;
delete TitleBarColor;
delete Font;
CleanUp();
} // end of ~QWindow()

// end of implementation of QWindow


